; Syntax: NASM, tab width = 8 spaces (or higher)
; - 8bitbubsy, 2023-2025
;
; WARNING: Doesn't call old interrupt 8 routine!
;
; Constants:
;  PIT_CLK      = 1193181.81818181 (recurring)
;  BYTES_PER_KB = 1024

	CPU 8086
	ORG 100h

NUM_WRITES	EQU 22 ; good compromise between bench time and accuracy
SCREEN_W	EQU 320
SCREEN_H	EQU 200

	mov ax,cs
	mov ds,ax
	; ---------------------
	mov ax,1600h	; check if we are running in DOS
	int 2Fh
	test al,al
	jz .L0
	lea dx,notDOSErrText
	jmp error
.L0:	; ---------------------
	call detectVGA
	jnc .L1
	lea dx,noVGAErrText
	jmp error
.L1:	; ---------------------
	; all good, let's get started
	; ---------------------
	mov ax,0F00h
	int 10h
	mov [oldVideoMode],al
	mov ax,13h
	int 10h
	; ---------------------
	call benchmarkVGA
	; ---------------------
	xor ah,ah
	mov al,[oldVideoMode]
	int 10h
	; ---------------------
	lea dx,thruputText
	call putString
	; ---------------------
	mov cx,[ticksHi]
	mov bx,[ticksLo]
	test cx,cx	; lo-clamp result to 1
	jnz .L2
	test bx,bx
	jnz .L2
	mov bx,1
.L2:	; ---------------------
	; AX:DX = (1640625000 + (Ticks/2)) / Ticks
	; 1640625000.0 = (PIT_CLK * SCREEN_W * SCREEN_H * NUM_WRITES) / BYTES_PER_KB
	; (perfectly round number)
	; ---------------------
	mov dx,cx
	mov ax,bx
	shr dx,1
	rcr ax,1
	add ax,0F368h	; lo-word of 1640625000
	adc dx,061C9h	; hi-word of 1640625000
	call divU32
	test dx,dx
	jnz .mbps
	cmp ax,1000
	jae .mbps
	; ---------------------
.kbps:	call printKbps
	jmp exitToDOS
.mbps:	; ---------------------
	; AX:DX /= 1024 (result clamped to 65535 and put in DI)
	; ---------------------
	push ax
	mov cx,10
.loop:	shr dx,1
	rcr ax,1
	loop .loop
	test dx,dx
	jz .L3
	mov ax,65535
.L3:	mov di,ax
	pop ax
	; ---------------------
	; get fractional part of AX:DX / 1024, and scale to 0..999
	; ---------------------
	mov cx,1024
	div cx
	mov ax,dx
	mov bx,1000
	mul bx
	mov bx,1024
	div bx
	; ---------------------
	mov bx,ax
	mov ax,di
	call printMbpsWithFrac
exitToDOS:
	mov ah,4Ch
	int 21h
error:
	call putString
	jmp exitToDOS

putString:
	mov ah,9
	int 21h
	ret

handleDigit:
	add al,'0'
	mov [di],al
	inc di
	xchg ax,dx
	xor dx,dx
	ret

printKbps:
	lea di,dec3Str
	mov bx,100
	div bx
	call handleDigit
	mov bx,10
	div bx
	call handleDigit
	call handleDigit
	; ---------------------
	lea di,dec3Str
	mov cx,3
.loop:	cmp byte [di],'0'
	jne .done
	inc di
	jmp .loop
.done:	mov dx,di
	call putString
	; ---------------------
	lea dx,endTextKbps
	call putString
	ret

	; AX = high part
	; BX = low part
printMbpsWithFrac:
	push bx
	xor dx,dx
	lea di,dec5Str
	mov bx,10000
	div bx
	call handleDigit
	mov bx,1000
	div bx
	call handleDigit
	mov bx,100
	div bx
	call handleDigit
	mov bx,10
	div bx
	call handleDigit
	call handleDigit
	; ---------------------
	lea di,dec5Str
	mov cx,5
.loop:	cmp byte [di],'0'
	jne .done
	inc di
	jmp .loop
.done:	mov dx,di
	call putString
	; ---------------------
	; handle fractional part
	; ---------------------
	lea dx,dotText
	call putString
	; ---------------------
	pop ax
	xor dx,dx
	lea di,dec3Str
	mov bx,100
	div bx
	call handleDigit
	mov bx,10
	div bx
	call handleDigit
	call handleDigit
	; ---------------------
	lea dx,dec3Str
	call putString
	; ---------------------
	lea dx,endTextMbps
	call putString
	ret

detectVGA:
	mov ax,1A00h
	int 10h
	cmp al,26
	jne .error
	cmp bl,1
	je .ok
	cmp bl,7
	je .ok
	cmp bl,8
	je .ok
	cmp bl,10
	je .ok
	cmp bl,11
	je .ok
	cmp bl,12
	je .ok
.error:	stc
	ret
.ok:	clc
	ret

	; Disassembled 32-bit division routine from Open Watcom V2's runtime.
	; github.com/open-watcom/open-watcom-v2/tree/master/bld/clib/cgsupp/a/i4d.asm
	;
	; Input:
	;  DX:AX = Dividend
	;  CX:BX = Divisor
	;
	; Output:
	;  DX:AX = Quotient
	;  CX:BX = Remainder
	;
divU32:
	test cx,cx
	jnz .L2
	dec bx
	jz .L1
	inc bx
	cmp bx,dx
	ja .L0
	mov cx,ax
	mov ax,dx
	xor dx,dx
	div bx
	xchg ax,cx
.L0:	div bx
	mov bx,dx
	mov dx,cx
	xor cx,cx
.L1:	ret
.L2:	cmp cx,dx
	jc .L4
	jnz .L3
	cmp bx,ax
	ja .L3
	sub ax,bx
	mov bx,ax
	xor cx,cx
	xor dx,dx
	mov ax,1
	ret
.L3:	xor cx,cx
	xor bx,bx
	xchg ax,bx
	xchg dx,cx
	ret
.L4:	push bp
	push si
	xor si,si
	mov bp,si
.L5:	add bx,bx
	adc cx,cx
	jc .L8
	inc bp
	cmp cx,dx
	jc .L5
	ja .L6
	cmp bx,ax
	jbe .L5
.L6:	clc
.L7:	adc si,si
	dec bp
	js .LB
.L8:	rcr cx,1
	rcr bx,1
	sub ax,bx
	sbb dx,cx
	cmc
	jc .L7
.L9:	add si,si
	dec bp
	js .LA
	shr cx,1
	rcr bx,1
	add ax,bx
	adc dx,cx
	jnc .L9
	jmp .L7
.LA:	add ax,bx
	adc dx,cx
.LB:	mov bx,ax
	mov cx,dx
	mov ax,si
	xor dx,dx
	pop si
	pop bp
	ret

pitOverflowInterrupt:
	inc word [ticksHi]
	jmp far [oldInt8]

%MACRO DO_FRAMEWRITE 0
	mov di,dx
	mov cx,bx
	cld
	rep stosw
%ENDMACRO

benchmarkVGA:
	cli
	; ---------------------
	xor ax,ax		; set new interrupt 8 vector
	mov es,ax
	mov ax,es:[32+0]	
	mov [oldInt8+0],ax
	mov ax,es:[32+2]
	mov [oldInt8+2],ax
	mov ax,pitOverflowInterrupt
	mov es:[32+0],ax
	mov es:[32+2],cs
	; ---------------------
	mov ax,0A000h		; prepare some stuff
	mov es,ax
	mov bx,(SCREEN_W*SCREEN_H)/2
	xor dx,dx		; MOV'd to DI for every rep stosw in benchmark
	mov [ticksHi],dx	; clear hi-ticks
	mov cx,2
	; ---------------------	
	mov al,34h		; set PIT channel 0 (mode 2, rate generator)
	out 43h,al
	xor ax,ax		; clear upper byte too, for DO_FRAMEWRITE maro
	out 40h,al
	out 40h,al
	; ---------------------
	lea si,ticksHi
	sti			; enable interrupts now
.wait:	cmp [si],cx		; wait until at least 2 PIT ints have fired (ticksHi >= 2)
	jb .wait
	; ---------------------
	; we're good to go, do benchmarking now
	; ---------------------
%REP NUM_WRITES
	DO_FRAMEWRITE
%ENDREP
	; --------------------- ; benching done
	cli
	; ---------------------	
	xor al,al		; read PIT channel 0
	out 43h,al
	in al,40h
	mov ah,al
	in al,40h
	xchg al,ah
	neg ax			; invert count (count down -> count up)
	mov [ticksLo],ax
	; ---------------------
	xor ax,ax		; set back old interrupt 8 vector
	mov es,ax
	mov ax,[oldInt8+0]	
	mov es:[32+0],ax
	mov ax,[oldInt8+2]
	mov es:[32+2],ax
	; ---------------------
	mov al,36h		; reset PIT channel 0 to BIOS default
	out 43h,al
	xor al,al
	out 40h,al
	out 40h,al
	; ---------------------
	sub word [ticksHi],2	; we waited 2 PIT interrupts before benchmark
	sti
	ret

	ALIGN 2
oldInt8:	DW 0
		DW 0
ticksLo:	DW 0
ticksHi:	DW 0
oldVideoMode:	DB 0
dec3Str:	DB "000",'$'
dec5Str:	DB "00000",'$'
dotText:	DB ".",'$'
notDOSErrText:	DB "ERROR: This program cannot be run from Windows! Please restart into DOS mode.",13,10,'$'
noVGAErrText:	DB "ERROR: This program requires a VGA graphics adapter.",13,10,'$'
thruputText:	DB "VGA thruput (16-bit writes): ",'$'
endTextKbps:	DB "kB/s",13,10,'$'
endTextMbps:	DB "MB/s",13,10,'$'
